print("import: 'menuinst'")
import menuinst

print("import: 'menuinst.api'")
import menuinst.api

print("import: 'menuinst._legacy'")
import menuinst._legacy

print("import: 'menuinst.platforms.win_utils.winshortcut'")
import menuinst.platforms.win_utils.winshortcut

